"""
sk custom locale file.
"""
from __future__ import annotations


translations = {
    # Relative time
    "ago": "pred {}",
    "from_now": "o {}",
    "after": "{0} po",
    "before": "{0} pred",
    # Date formats
    "date_formats": {
        "LTS": "HH:mm:ss",
        "LT": "HH:mm",
        "LLLL": "dddd, D. MMMM YYYY HH:mm",
        "LLL": "D. MMMM YYYY HH:mm",
        "LL": "D. MMMM YYYY",
        "L": "DD.MM.YYYY",
    },
}
